import numpy as np
from distutils.version import LooseVersion
import warnings

import keras
import keras.backend as K
from tensorflow.keras.layers import Input, Conv3D, MaxPooling3D, Flatten, Dense
from tensorflow.keras.layers import GlobalAveragePooling3D, BatchNormalization, Activation, Add

from tensorflow.keras.models import Model
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.callbacks import ReduceLROnPlateau, TensorBoard

from csbdeep.utils import _raise, axes_check_and_normalize, normalize
from csbdeep.models import BaseConfig, BaseModel

from .psf import PsfGenerator3D
from .zernike import random_zernike_wavefront, ensure_dict
from .noise import add_random_noise
from .phantoms import Phantom3D
from .utils import cropper3D

from scipy.signal import convolve
from scipy.ndimage.filters import gaussian_filter

# TODO
# try:
#     from gputools import fft_convolve
#     gputools_flag=False
# except:
#     gputools_flag=False
# gputools_flag = False

# print(f"GPU Tools {gputools_flag}")

class Data:

    def __init__(self,
                 amplitude_ranges, order='noll', normed=True,
                 batch_size=1,
                 psf_shape=(64,64,64), units=(0.1,0.1,0.1), na_detection=1.1, lam_detection=.5, n=1.33, n_threads=4,
                 noise_snr=None, noise_mean=None, noise_sigma=None, noise_perlin_flag=False, gaussian_blur_sigma=None,
                 phantom_params=None,
                 crop_shape=None, jitter=False, max_jitter=None,
                 planes=None
                 ):
        """Encapsulates data generation

        :param psf_shape: tuple, shape of psf, eg (32,32,32)
        :param units: tuple, units in microns, eg (0.1,0.1,0.1)
        :param lam_detection: scalar, wavelength in micrometer, eg 0.5
        :param n: scalar, refractive index, eg 1.33
        :param na_detection: scalar, numerical aperture of detection objective, eg 1.1
        :param n_threads: integer, for multiprocessing
        :param noise_snr: scalar or tuple, signal to noise ratio
        :param noise_mean: scalar or tuple, mean background noise
        :param noise_sigma: scalar or tuple, sigma for Gaussian noise
        :param noise_perlin_flag: boolean, default is False
        :param gaussian_blur_sigma: float, sigma for Gaussian blurring after adding noise, default is None
        :param phantom_params : dictionary, phantom name and parameters for that phantom
        :param crop_shape: tuple, crop shape
        :param jitter: Boolean, randomly move the center point within a given limit, default is False
        :param max_jitter: tuple, maximum displacement for jitter, if None then it gets a default value
        :param planes: list, z planes with respect to center, if None then it takes all the planes
        """

        self.psfgen = PsfGenerator3D(psf_shape=psf_shape, units=units, lam_detection=lam_detection, n=n, na_detection=na_detection, n_threads=n_threads)
        self.order = order
        self.normed = normed
        self.amplitude_ranges = ensure_dict(amplitude_ranges, order)
        self.batch_size = batch_size
        self.snr = noise_snr
        self.sigma = noise_sigma
        self.mean = noise_mean
        self.noise_perlin_flag = noise_perlin_flag
        self.gaussian_blur_sigma = gaussian_blur_sigma
        if planes is not None:
            self.planes = np.array(planes)
            if crop_shape is None:
                self.crop_shape = tuple((self.planes.shape[0],psf_shape[1],psf_shape[2]))
            else:
                self.crop_shape = tuple((self.planes.shape[0],crop_shape[1],crop_shape[2]))
        else:
            self.planes = planes
            self.crop_shape = crop_shape
        self.jitter = jitter
        self.max_jitter = max_jitter
        self.phantom_params = phantom_params
        if self.phantom_params is not None:
            self.phantom_params.setdefault('shape', psf_shape)
            self.phantom_params.setdefault('units', units)
            self.phantom = Phantom3D.instantiate(**self.phantom_params)
        else:
            self.phantom = None


    def _single_psf(self):
        phi = random_zernike_wavefront(self.amplitude_ranges, order=self.order)
        psf = self.psfgen.incoherent_psf(phi, normed=self.normed)


        if self.phantom is not None:
            self.phantom.generate()
            obj =  self.phantom.get()
            # if gputools_flag:
            #     psf = fft_convolve(obj, psf) #the center is shifted by 1 pixel as compared to convolve
            # else:
            psf = convolve(obj, psf, 'same')

        if self.snr is not None and self.sigma is not None and self.mean is not None:
            self.noise_flag = True
            psf = add_random_noise(psf, self.snr, self.mean, self.sigma, perlin=self.noise_perlin_flag)
            if self.gaussian_blur_sigma is not None:
                gaussian_blur = (self.gaussian_blur_sigma, self.gaussian_blur_sigma) if np.isscalar(self.gaussian_blur_sigma) else self.gaussian_blur_sigma
                gaussian_blur = np.random.uniform(*gaussian_blur)
                psf = gaussian_filter(psf,gaussian_blur)
        else:
            self.noise_flag = False
            if self.snr is not None or self.sigma is not None or self.mean is not None:
                warnings.warn("No noise added")

        if self.crop_shape is not None:
            self.crop_flag = True
            psf = cropper3D(psf, self.crop_shape, jitter=self.jitter, max_jitter=self.max_jitter, planes=self.planes)
        else:
            self.crop_flag =  False

        return psf, phi.amplitudes_requested


    def generator(self):
        while True:
            psfs, amplitudes = zip(*(self._single_psf() for _ in range(self.batch_size)))
            psfs = [normalize(psf) for psf in psfs]
            X = np.expand_dims(np.stack(psfs, axis=0), -1)
            Y = np.stack(amplitudes, axis=0)
            yield X, Y



class Config(BaseConfig):
    """Configuration for phasenet models

    :param zernike_amplitude_ranges: dictionary or list, the values should either a scalar indicating the absolute magnitude
            or a tuple with upper and lower bound, default is {'vertical coma': (-0.2,0.2)}
    :param zernike_order: string, Zernike nomenclature used when the amplitude ranges are given as a list, default is 'noll'
    :param zernike_normed: Boolean, whether the Zernike are normalized according, default is True
    :param net_architecture: convnet or resnet, default is convnet
    :param net_kernel_size: convolution kernel size, default is (3,3,3)
    :param net_pool_size: max pool kernel size, default is (1,2,2)
    :param net_activation: activation, default is 'tanh'
    :param net_padding: padding for convolution, default is 'same'
    :param psf_shape: tuple, shpae of the psf, default is (64,64,64)
    :param psf_units: tuple, voxel unit (z,y,x) in um, default is (0.1,0.1,0.1)
    :param psf_na_detection: scalar, numerical aperture default is 1.1
    :param psf_lam_detection: scalar, wavelength in um, default is 0.5
    :param psf_n: scalar, refractive index of immersion medium, default is 1.33
    :param noise_snr: scalar or tuple, signal to noise ratio
    :param noise_mean: scalar or tuple, mean background noise
    :param noise_sigma: scalar or tuple, sigma for Gaussian noise
    :param gaussian_blur_sigma: float, sigma for Gaussian blurring after adding noise, default is None
    :param phantom_params: dictionary, parameters for the chosen phantom, e.g. {'name':'sphere','radius':0.1}
    :param crop_shape: tuple, crop shape
    :param jitter: Boolean, randomly move the center point within a given limit, default is False
    :param max_jitter: tuple, maximum displacement for jitter, if None then it gets a default value
    :param train_loss: string, training loss, default is 'mse'
    :param train_epochs: integer, number of epochs for training, default is 400
    :param train_steps_per_epoch: integer, number of steps per epoch, default is 5
    :param train_learning_rate: scalar, leaning rate, default is 0.0003
    :param train_batch_size: integer, batch size for training the network, default is 8
    :param train_n_val: integer, number of validation data, default is 128
    :param train_tensorboard: boolean, create tensor-board, default is True
    :param planes: list, z planes with respect to center, if None then it takes all the planes

    """

    def __init__(self, axes='ZYX', n_channel_in=1, **kwargs):
        """See class docstring."""

        super().__init__(axes=axes, n_channel_in=n_channel_in, n_channel_out=1)

        # directly set by parameters
        # ...

        # default config (can be overwritten by kwargs below)
        self.zernike_amplitude_ranges  = {'vertical coma': (-0.2,0.2)}
        self.zernike_order             = 'noll'
        self.zernike_normed            = True

        self.net_architecture         = "convnet"
        self.net_kernel_size           = (3,3,3)
        self.net_pool_size             = (1,2,2)
        self.net_activation            = 'tanh'
        self.net_padding               = 'same'

        self.psf_shape                 = (64,64,64)
        self.psf_units                 = (0.1,0.1,0.1)
        self.psf_na_detection          = 1.1
        self.psf_lam_detection         = 0.5
        self.psf_n                     = 1.33
        self.noise_mean                = 100
        self.noise_sigma               = 3.5
        self.noise_snr                 = (1.,5)
        self.noise_perlin_flag         = False
        self.gaussian_blur_sigma       = None
        self.phantom_params            = {'name':'points', 'num':1}
        self.crop_shape                = (32,32,32)
        self.jitter                    = True
        self.max_jitter                = None
        self.planes                    = None

        self.train_loss                = 'mse'
        self.train_epochs              = 400
        self.train_steps_per_epoch     = 5
        self.train_learning_rate       = 0.0003
        self.train_batch_size          = 8
        self.train_n_val               = 128
        self.train_tensorboard         = True
        
        # remove derived attributes that shouldn't be overwritten
        for k in ('n_dim', 'n_channel_out'):
            try: del kwargs[k]
            except KeyError: pass

        self.update_parameters(False, **kwargs)

        self.n_channel_out = len(random_zernike_wavefront(self.zernike_amplitude_ranges))


class PhaseNet(BaseModel):
    """PhaseNet model.

    Parameters
    ----------
    config : :class:`Config` or None
        Will be saved to disk as JSON (``config.json``).
        If set to ``None``, will be loaded from disk (must exist).
    name : str or None
        Model name. Uses a timestamp if set to ``None`` (default).
    basedir : str
        Directory that contains (or will contain) a folder with the given model name.

    Raises
    ------
    FileNotFoundError
        If ``config=None`` and config cannot be loaded from disk.
    ValueError
        Illegal arguments, including invalid configuration.

    Attributes
    ----------
    config : :class:`Config`
        Configuration, as provided during instantiation.
    keras_model : `Keras model <https://keras.io/getting-started/functional-api-guide/>`_
        Keras neural network model.
    name : str
        Model name.
    logdir : :class:`pathlib.Path`
        Path to model folder (which stores configuration, weights, etc.)
    """

    @property
    def _config_class(self):
        return Config


    @property
    def _axes_out(self):
        return 'C'

    def get_model_input_shape(self):
        if self.config.planes is not None:
            _p = np.array(self.config.planes)
            if self.config.crop_shape is None:
                config.model_input_shape = tuple((_p.shape[0],self.config.psf_shape[1],self.config.psf_shape[2]))
            else:
                model_input_shape = tuple((_p.shape[0],self.config.crop_shape[1],self.config.crop_shape[2]))
        elif self.config.crop_shape is not None:
            model_input_shape = self.config.crop_shape
        else :
            model_input_shape = self.config.psf_shape
        return model_input_shape

    def _build(self):
        model_input_shape = self.get_model_input_shape()
        input_shape = tuple(model_input_shape) + (self.config.n_channel_in,)
        output_size = self.config.n_channel_out
        kernel_size = self.config.net_kernel_size
        pool_size = self.config.net_pool_size
        activation = self.config.net_activation
        padding = self.config.net_padding

        if self.config.net_architecture == 'resnet':
            return self._resnet(input_shape, output_size, kernel_size, activation, padding, pool_size)
        elif self.config.net_architecture == 'convnet':
            return self._convnet(input_shape, output_size, kernel_size, activation, padding, pool_size)
        else:
            _raise(ValueError("Network architecture not defined"))

    def _convnet(self, input_shape, output_size, kernel_size, activation, padding, pool_size):

        print(">>> input_shape", input_shape)
        print(">>> output_size", output_size)
        print(">>> kernel_size", kernel_size)
        print(">>> activation", activation)
        print(">>> padding", padding)
        print(">>> pool_size", pool_size)

        inp = Input(name='X', shape=input_shape)
        t = Conv3D(8, name='conv1', kernel_size=kernel_size, activation=activation, padding=padding)(inp)
        t = Conv3D(8, name='conv2', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = MaxPooling3D(name='maxpool1', pool_size=pool_size)(t)
        t = Conv3D(16, name='conv3', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = Conv3D(16, name='conv4', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = MaxPooling3D(name='maxpool2', pool_size=pool_size)(t)
        t = Conv3D(32, name='conv5', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = Conv3D(32, name='conv6', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = MaxPooling3D(name='maxpool3', pool_size=pool_size)(t)
        t = Conv3D(64, name='conv7', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = Conv3D(64, name='conv8', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = MaxPooling3D(name='maxpool4', pool_size=pool_size)(t)
        t = Conv3D(128, name='conv9', kernel_size=kernel_size, activation=activation, padding=padding)(t)
        t = Conv3D(128, name='conv10', kernel_size=kernel_size, activation=activation, padding=padding)(t)

        if input_shape[0] == 1:
            t = MaxPooling3D(name='maxpool5', pool_size=(1, 2, 2))(t)
        else:
            t = MaxPooling3D(name='maxpool5', pool_size=(2, 2, 2))(t)
        t = Flatten(name='flat')(t)
        t = Dense(64, name='dense1', activation=activation)(t)
        t = Dense(64, name='dense2', activation=activation)(t)

        oup = Dense(output_size, name='Y', activation='linear')(t)

        return Model(inputs=inp, outputs=oup)


    def _resnet(self, input_shape, output_size, kernel_size, activation, padding, pool_size):

        def resnet_block(n_filters, kernel_size=kernel_size, batch_norm=True, downsample=False, kernel_initializer="he_normal"):
            def f(inp):
                strides = (2, 2, 2) if downsample else (1, 1, 1)
                x = Conv3D(n_filters, kernel_size, padding='same', use_bias=(not batch_norm),
                           kernel_initializer=kernel_initializer, strides=strides)(inp)
                if batch_norm:
                    x = BatchNormalization()(x)
                x = Activation(activation)(x)

                x = Conv3D(n_filters, kernel_size, padding=padding, use_bias=(not batch_norm),kernel_initializer=kernel_initializer)(x)
                if batch_norm:
                    x = BatchNormalization()(x)

                if downsample:
                    inp = Conv3D(n_filters, (1, 1, 1), padding=padding, use_bias=(not batch_norm), kernel_initializer=kernel_initializer,
                                 strides=strides)(inp)
                    if batch_norm:
                        inp = BatchNormalization()(inp)

                x = Add()([inp, x])
                x = Activation(activation)(x)
                return x

            return f

        inp = Input(input_shape, name='X')
        x = inp
        n = 16

        depth = 3
        for i in range(depth):
            x = resnet_block(n * (2 ** i), downsample=(i > 0))(x)
            x = resnet_block(n * (2 ** i))(x)
        x = GlobalAveragePooling3D()(x)
        oup = Dense(output_size, name='Y')(x)

        return Model(inp, oup)


    def prepare_for_training(self, optimizer=None):
        """Prepare for neural network training.

        Compiles the model and creates
        `Keras Callbacks <https://keras.io/callbacks/>`_ to be used for training.

        Note that this method will be implicitly called once by :func:`train`
        (with default arguments) if not done so explicitly beforehand.

        Parameters
        ----------
        optimizer : obj or None
            Instance of a `Keras Optimizer <https://keras.io/optimizers/>`_ to be used for training.
            If ``None`` (default), uses ``Adam`` with the learning rate specified in ``config``.

        """
        if optimizer is None:
            optimizer = Adam(lr=self.config.train_learning_rate)

        self.keras_model.compile(optimizer, loss=self.config.train_loss)

        self.callbacks = []
        if self.basedir is not None:
            self.callbacks += self._checkpoint_callbacks()

            if self.config.train_tensorboard:
                self.callbacks.append(TensorBoard(log_dir=str(self.logdir), write_graph=False))

        self._model_prepared = True


    def train(self, seed=None, epochs=None, steps_per_epoch=None):
        """Train the neural network with the given data.

        Parameters
        ----------
        seed : int
            Convenience to set ``np.random.seed(seed)``. (To obtain reproducible validation patches, etc.)
        epochs : int
            Optional argument to use instead of the value from ``config``.
        steps_per_epoch : int
            Optional argument to use instead of the value from ``config``.

        Returns
        -------
        ``History`` object
            See `Keras training history <https://keras.io/models/model/#fit>`_.

        """
        if seed is not None:
            # https://keras.io/getting-started/faq/#how-can-i-obtain-reproducible-results-using-keras-during-development
            np.random.seed(seed)
        if epochs is None:
            epochs = self.config.train_epochs
        if steps_per_epoch is None:
            steps_per_epoch = self.config.train_steps_per_epoch

        if not self._model_prepared:
            self.prepare_for_training()

        data_kwargs = dict (
            amplitude_ranges     = self.config.zernike_amplitude_ranges,
            order                = self.config.zernike_order,
            normed               = self.config.zernike_normed,
            psf_shape            = self.config.psf_shape,
            units                = self.config.psf_units,
            na_detection         = self.config.psf_na_detection,
            lam_detection        = self.config.psf_lam_detection,
            n                    = self.config.psf_n,
            noise_snr            = self.config.noise_snr,
            noise_mean           = self.config.noise_mean,
            noise_sigma          = self.config.noise_sigma,
            noise_perlin_flag    = self.config.noise_perlin_flag,
            gaussian_blur_sigma  = self.config.gaussian_blur_sigma,
            phantom_params       = self.config.phantom_params,
            crop_shape           = self.config.crop_shape,
            jitter               = self.config.jitter,
            max_jitter           = self.config.max_jitter,
            planes               = self.config.planes,
        )

        # generate validation data and store in numpy arrays
        data_val = Data(batch_size=self.config.train_n_val, **data_kwargs) # TODO: turn augmentation off here
        data_val = next(data_val.generator())

        data_train = Data(batch_size=self.config.train_batch_size, **data_kwargs)

        history = self.keras_model.fit_generator(generator=data_train.generator(), validation_data=data_val,
                                                 epochs=epochs, steps_per_epoch=steps_per_epoch,
                                                 callbacks=self.callbacks, verbose=1)
        self._training_finished()
        return history


    def predict(self, img, axes=None, normalizer=None, **predict_kwargs):
        """Predict.

        Parameters
        ----------
        img : :class:`numpy.ndarray`
            Input image
        axes : str or None
            Axes of the input ``img``.
            ``None`` denotes that axes of img are the same as denoted in the config.
        normalizer : :class:`csbdeep.data.Normalizer` or None
            (Optional) normalization of input image before prediction.
            Note that the default (``None``) assumes ``img`` to be already normalized.
        predict_kwargs: dict
            Keyword arguments for ``predict`` function of Keras model.

        """
        if axes is None:
            axes = self.config.axes
            assert 'C' in axes
            if img.ndim == len(axes)-1 and self.config.n_channel_in == 1:
                # img has no dedicated channel axis, but 'C' always part of config axes
                axes = axes.replace('C','')

        axes     = axes_check_and_normalize(axes,img.ndim)
        axes_net = self.config.axes

        _permute_axes = self._make_permute_axes(axes, axes_net)
        x = _permute_axes(img) # x has axes_net semantics
        x.shape == tuple(self.get_model_input_shape()) + (self.config.n_channel_in,) or _raise(ValueError())

        normalizer = self._check_normalizer_resizer(normalizer, None)[0]
        x = normalizer.before(x, axes_net)

        return self.keras_model.predict(x[np.newaxis], **predict_kwargs)[0]



import numpy as np
from csbdeep.utils import _raise



class PsfGenerator3D:
    """
    3D PSF generator, courtesy of Martin Weigert (https://github.com/maweigert)
    """

    def __init__(self, psf_shape, units, lam_detection, n, na_detection, n_threads=4):

        """
        encapsulates 3D PSF generator

        :param psf_shape: tuple, psf shape as (z,y,x), e.g. (64,64,64)
        :param units: tuple, voxel size in microns, e.g. (0.1,0.1,0.1)
        :param lam_detection: scalar, wavelength in microns, e.g. 0.5
        :param n: scalar, refractive index, eg 1.33
        :param na_detection: scalar, numerical aperture of detection objective, eg 1.1
        :param n_threads: integer, for multiprocessing
        """

        psf_shape = tuple(psf_shape)
        units = tuple(units)

        # setting up the fourier domain and the wavefronts...

        self.Nz, self.Ny, self.Nx = psf_shape
        self.dz, self.dy, self.dx = units

        self.na_detection = na_detection
        self.lam_detection = lam_detection

        self.n = n

        kx = np.fft.fftfreq(self.Nx, self.dx)
        ky = np.fft.fftfreq(self.Ny, self.dy)

        z = self.dz * (np.arange(self.Nz) - self.Nz // 2)

        self.KZ3, self.KY3, self.KX3 = np.meshgrid(z, ky, kx, indexing="ij")
        KR3 = np.sqrt(self.KX3 ** 2 + self.KY3 ** 2)

        # the cutoff in fourier domain
        self.kcut = 1. * na_detection / self.lam_detection
        self.kmask3 = (KR3 <= self.kcut)

        H = np.sqrt(1. * self.n ** 2 - KR3 ** 2 * lam_detection ** 2)

        self._H = H

        out_ind = np.isnan(H)
        self.kprop = np.exp(-2.j * np.pi * self.KZ3 / lam_detection * H)
        self.kprop[out_ind] = 0.

        self.kbase = self.kmask3 * self.kprop

        KY2, KX2 = np.meshgrid(ky, kx, indexing="ij")
        KR2 = np.hypot(KX2, KY2)

        self.krho = KR2 / self.kcut
        self.kphi = np.arctan2(KY2, KX2)
        self.kmask2 = (KR2 <= self.kcut)

        self.myzifftn = lambda x: np.fft.ifftn(x,axes=(1,2))


    def masked_phase_array(self, phi, normed=True):
        """
        returns masked Zernike polynomial for back focal plane, masked according to the setup

        :param phi: Zernike/ZernikeWavefront object
        :param normed: boolean, multiplied by normalization factor, eg True
        :return: masked wavefront, 2d array
        """
        return self.kmask2 * phi.phase(self.krho, self.kphi, normed=normed, outside=None)


    def coherent_psf(self, phi, normed=True):
        """
        returns the coherent psf for a given wavefront phi

        :param phi: Zernike/ZernikeWavefront object
        :return: coherent psf, 3d array
        """
        phi = self.masked_phase_array(phi, normed=normed)
        ku = self.kbase * np.exp(2.j * np.pi * phi / self.lam_detection)
        res = self.myzifftn(ku)
        return np.fft.fftshift(res, axes=(0,))

    def incoherent_psf(self, phi, normed=True):
        """
        returns the incoherent psf for a given wavefront phi
           (which is just the squared absolute value of the coherent one)
           The psf is normalized such that the sum intensity on each plane equals one

        :param phi: Zernike/ZernikeWavefront object
        :return: incoherent psf, 3d array
        """
        _psf = np.abs(self.coherent_psf(phi, normed=normed)) ** 2
        _psf = np.array([p/np.sum(p) for p in _psf])
        return np.fft.fftshift(_psf)





import numpy
import h5py

def get_wofry_data(root, dir_out="./", training_ratio=2/3, verbose=1, gs_or_z=0, nbin=1):
    if gs_or_z == 0:
        filename = "%s%s_targets_gs.txt" % (dir_out, root)
    else:
        filename = "%s%s_targets_z.txt" % (dir_out, root)

    tmp = numpy.loadtxt(filename)

    targets = tmp[:,1:].copy()

    if verbose: print(targets.shape, targets)

    h5_file = "%s%s_block.h5" % (dir_out, root)

    f = h5py.File(h5_file, 'r')

    data = f['allsamples/intensity/stack_data'][:]

    f.close()

    if verbose: print(data.shape, data)

    size_data = data.shape[0]
    size_target = targets.shape[0]

    if size_data != size_target:
        raise Exception("Data and targets must have the same size.")


    istart_training = 0
    iend_training = int(training_ratio * size_data)
    istart_test = iend_training
    iend_test = size_data


    if nbin > 0: # normal binning
        return (data[istart_training:iend_training,:,::nbin].copy(), targets[istart_training:iend_training].copy()), \
               (data[istart_test:iend_test,:,::nbin].copy(), targets[istart_test:iend_test].copy())
    else: # take the last part
        return (data[istart_training:iend_training,:,(data.shape[-1]//numpy.abs(nbin)):].copy(), targets[istart_training:iend_training].copy()), \
           (data[istart_test:iend_test,:,(data.shape[-1]//numpy.abs(nbin)):].copy(), targets[istart_test:iend_test].copy())

if __name__ == "__main__":


    dir_out = "/users/srio/Oasys/ML_TRAIN2/"  # where the results are going to be written
    root = "tmp_ml"


    (training_data, training_target), (test_data, test_target) = get_wofry_data(root, dir_out=dir_out, nbin=1)

    print("Training [data/target]: ", training_data.shape, training_target.shape)
    print("Test: [data/target]", test_data.shape, test_target.shape)





import numpy
from mlcrl.phasenet.zernike import Zernike
from srxraylib.plot.gol import plot_table


def create_1d_zernike_basis(size=512, noll=[6, 8, 10, 11, 12, 14, 22, 37], width=2.0, filename=None, do_plot=False):

    nnoll = len(noll)

    basis = numpy.zeros((size, len(noll)))
    basis_x = numpy.linspace(-0.5*width, 0.5*width, size)
    for i in range(len(noll)):
        z = Zernike(noll[i], order='noll')
        w = z.polynomial_vertical(size)
        basis[:, i] = w

    if do_plot: plot_table(basis_x, basis.T, xtitle="position [um]", ytitle="basis",
               title="non-orthogonal basis (noll-indexed Vertical cut of Zernike polynomials)", legend=noll)

    if filename is not None:  # write to file
        f = open(filename, 'w')
        for i in range(size):
            f.write("%g  " % (basis_x[i]))
            for j in range(nnoll):
                f.write("%12.10f  " % (basis[i, j]))
            f.write("\n")
        f.close()
        print("File %s written to disk." % filename)

    return basis_x, basis



from mlcrl.phasenet.zernike import Zernike
import numpy
from srxraylib.plot.gol import plot, plot_table
import matplotlib.pyplot as plt

def create_1d_zernike_sampled_profiles(nsamples,
                    noll=[6, 8, 10, 11, 12, 14, 22, 37],
                    distrubution=['n','n','n','n','n','n','n','n'],
                    scale=[1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],
                    size=128,
                    factor=1.0,
                    seed=69,
                    do_plot=False):

    Y = numpy.zeros((size, nsamples))
    C = numpy.zeros((len(noll), nsamples))

    rg = numpy.random.default_rng(seed)
    x = numpy.linspace(-1, 1, size)

    # noll       = [6,     8,  10,  11,  12,  14,  22, 37]
    # distrubution = ['n', 'n', 'n', 'u', 'n', 'n', 'u', 'u']
    # scale        = [0.5, 0.5, 0.5, 2.3, .05, .05, 1.0, 0.5]

    for i in range(nsamples):
        y = numpy.zeros_like(x)

        for ij, j in enumerate(noll):
            z = Zernike(j, order='noll')

            if distrubution[ij] == 'n':
                c = rg.normal(loc=0, scale=scale[ij] * factor)
            elif distrubution[ij] == 'u':
                c = rg.uniform(-scale[ij] * factor, scale[ij] * factor)
            else:
                raise Exception("Why am I here?")

            # if j <= 10:
            #     # Z5 (Zcoeffs[4]) to Z10 (Zcoeffs[9]) are more important and have a higher weight
            #     c = rg.normal(loc=0, scale=0.5 * factor)
            # elif j == 11:
            #     # Spherical aberration 1st order (Z11)
            #     c = rg.uniform(-2.3 * factor, 2.3 * factor)
            # elif j >= 12 and j <= 21:
            #     # Z12 (Zcoeffs[11]) to Z21 (Zcoeffs[20]) and Z23 (Zcoeffs[22]) to Z36 (Zcoeffs[35]) are very low
            #     c = rg.normal(loc=0, scale=0.05 * factor)
            # elif j == 22:
            #     # Spherical aberration 2nd order (Z22)
            #     c = rg.uniform(-1. * factor, 1. * factor)
            # elif j == 37:
            #     # Spherical aberration 3nd order (Z37)
            #     c = rg.uniform(-0.5 * factor, 0.5 * factor)
            # else:
            #     raise Exception("Why am I here?")

            w = z.polynomial_vertical(size)
            y += c * w
            C[ij, i] = c
        Y[:, i] = y

        if do_plot: plot(x, y, title="sampled # %s" % (i + 1))

    return C, Y

if __name__ == "__main__":



    # For visualizing the Zernikes we define the Zernike object and call `polynomial()`.
    # The size of the 2D array is passed as a parameter

    if False:
        fig, ax = plt.subplots(4,10, figsize=(16,10))
        z_old = Zernike(1, order='noll').polynomial(128)
        for i,a in enumerate(ax.ravel()):
            z = Zernike(i+1, order='noll')
            w = z.polynomial(128)
            print(w.shape)
            a.imshow(w)
            if z.name is not None:
                a.set_title(z.name)
                # a.set_title("Noll: %d (%d,%d)" % (i + 1, z.n, z.m))
            else:
                a.set_title("Noll: %d (%d,%d)" % (i+1, z.n, z.m))
            a.axis('off')
            # print(">>>>>> i.(i-1) = ", numpy.nansum((z_old * w)) )

        plt.show()


    #
    # check normalization 2D
    #
    if False:
        noll = [6, 8, 10, 11, 12, 14, 22, 37]
        nnoll = len(noll)
        size = 128 * 4
        for i in range(nnoll):
            for j in range(nnoll):
                z1 = Zernike(noll[i], order='noll')
                z2 = Zernike(noll[j], order='noll')
                tmp = numpy.nansum(z1.polynomial(size=size) * z2.polynomial(size=size))/size**2
                if numpy.abs(tmp) < 1e-12:
                    tmp = 0
                print(">>> norm noll %d  %d: %g" % (noll[i],noll[j], tmp))

    #
    # check NON-normalization 1D
    #
    if False:
        size = 128 * 4
        noll = [6, 8, 10, 11, 12, 14, 22, 37]
        nnoll = len(noll)
        for i in range(nnoll):
            for j in range(nnoll):
                z1 = Zernike(noll[i], order='noll')
                z2 = Zernike(noll[j], order='noll')
                tmp = numpy.nansum(z1.polynomial_vertical(size=size) * z2.polynomial_vertical(size=size))/size
                if numpy.abs(tmp) < 1e-12:
                    tmp = 0
                print(">>> norm V profile noll %d  %d: %g" % (noll[i],noll[j], tmp))


    #
    # retrieve and plot 1d (vertical) Zernike profiles
    #

    if False:
        size = 128
        noll = [6, 8, 10, 11, 12, 14, 22, 37]

        fig, ax = plt.subplots(3,3, figsize=(10,10))
        A = ax.ravel()

        j = -1
        for i in noll:
            z = Zernike(i, order='noll')
            w = z.polynomial_vertical(size)
            print(i,z.name, z.n, z.m)
            if True: # z.m >= 0:
                j += 1
                a = A[j]
                a.plot(numpy.linspace(-1,1,size), w) #
            a.set_title("%s (%d,%d)" % (z.name, z.n, z.m))
            z_old = w
        plt.show()



    #
    # sampled polynomials
    #
    if True:
        size = 128
        nsamples = 20
        seed = 69  # seed for generation of the random Zernike profiles
        noll         =             [6,     8,  10,  11,  12,  14,  22,  37]
        distrubution =             ['n', 'n', 'n', 'u', 'n', 'n', 'u', 'u']
        scale        = numpy.array([0.5, 0.5, 0.5, 2.3, .05, .05, 1.0, 0.5]) * 1e-6
        width = 1500e-6

        C, Y  = create_1d_zernike_sampled_profiles(nsamples,
                                                   noll=noll, distrubution=distrubution, scale=scale, factor=5.0,
                                                   size=size, seed=seed, do_plot=0)


        print(Y.shape,C)
        xx = numpy.linspace(-width/2,width/2,size)
        plot_table(xx*1e6, Y.T*1e6, xtitle="Lens position [um]", ytitle="Sampled thickness [um]", title="sampled profiles",
                   legend=numpy.arange(nsamples))

        #
        # write files
        #
        dir = "./"
        root = "tmp_ml"

        for i in range(nsamples):
            #txt
            filename = "%s%s%04d.txt" % (dir, root, i+1)
            f = open(filename, 'w')

            f.write("# noll ")
            for j in range(len(noll)):
                f.write("%d  " % noll[j])
            f.write("\n")

            f.write("# coeff ")
            for j in range(len(noll)):
                f.write("%g  " % C[j,i])
            f.write("\n")

            print("File written to disk: %s" % filename)

            # dat
            filename = "%s%s%04d.dat" % (dir, root, i + 1)
            f = open(filename, 'w')

            for j in range(size):
                f.write("%g   %g\n" % (xx[j], Y[j,i]))
            f.close()
            print("File written to disk: %s" % filename)


        # overwrite 000 with no deformation
        # txt - no deformation
        filename = "%s%s%04d.txt" % (dir, root, 0)
        f = open(filename, 'w')

        f.write("# noll ")
        for j in range(len(noll)):
            f.write("%d  " % noll[j])
        f.write("\n")

        f.write("# coeff ")
        for j in range(len(noll)):
            f.write("%g  " % 0.0)
        f.write("\n")

        print("File written to disk: %s" % filename)

        # dat - no deformation
        filename = "%s%s%04d.dat" % (dir, root, 0)
        f = open(filename, 'w')

        for j in range(size):
            f.write("%g   %g\n" % (xx[j], 0.0))
        f.close()
        print("File written to disk: %s" % filename)



import numpy as np
from scipy.special import binom
from csbdeep.utils import _raise
from functools import lru_cache



def nm_to_noll(n, m):
    j = (n*(n+1))//2 + abs(m)
    if m> 0 and n%4 in (0,1): return j
    if m< 0 and n%4 in (2,3): return j
    if m>=0 and n%4 in (2,3): return j+1
    if m<=0 and n%4 in (0,1): return j+1
    assert False


def nm_to_ansi(n, m):
    return (n*(n+2) + m) // 2


def nm_normalization(n, m):
    """the norm of the zernike mode n,m in born/wolf convetion

    i.e. sqrt( \int | z_nm |^2 )
    """
    return np.sqrt((1.+(m==0))/(2.*n+2))


def nm_polynomial(n, m, rho, theta, normed=True):
    """returns the zernike polyonimal by classical n,m enumeration

    if normed=True, then they form an orthonormal system

        \int z_nm z_n'm' = delta_nn' delta_mm'

        and the first modes are

        z_nm(0,0)  = 1/sqrt(pi)*
        z_nm(1,-1) = 1/sqrt(pi)* 2r cos(phi)
        z_nm(1,1)  = 1/sqrt(pi)* 2r sin(phi)
        z_nm(2,0)  = 1/sqrt(pi)* sqrt(3)(2 r^2 - 1)
        ...
        z_nm(4,0)  = 1/sqrt(pi)* sqrt(5)(6 r^4 - 6 r^2 +1)
        ...

    if normed =False, then they follow the Born/Wolf convention
        (i.e. min/max is always -1/1)

        \int z_nm z_n'm' = (1.+(m==0))/(2*n+2) delta_nn' delta_mm'

        z_nm(0,0)  = 1
        z_nm(1,-1) = r cos(phi)
        z_nm(1,1)  =  r sin(phi)
        z_nm(2,0)  = (2 r^2 - 1)
        ...
        z_nm(4,0)  = (6 r^4 - 6 r^2 +1)


    """
    if abs(m) > n:
        raise ValueError(" |m| <= n ! ( %s <= %s)" % (m, n))

    if (n - m) % 2 == 1:
        return 0 * rho + 0 * theta

    radial = 0
    m0 = abs(m)

    for k in range((n - m0) // 2 + 1):
        radial += (-1.) ** k * binom(n - k, k) * binom(n - 2 * k, (n - m0) // 2 - k) * rho ** (n - 2 * k)

    radial *= (rho <= 1.)

    if normed:
        prefac = 1. / nm_normalization(n, m)
    else:
        prefac = 1.
    if m >= 0:
        return prefac * radial * np.cos(m0 * theta)
    else:
        return prefac * radial * np.sin(m0 * theta)


@lru_cache(maxsize=32)
def rho_theta(size):
    r = np.linspace(-1,1,size)
    X,Y = np.meshgrid(r,r, indexing='ij')
    rho = np.hypot(X,Y)
    theta = np.arctan2(Y,X)
    return rho, theta


@lru_cache(maxsize=32)
def outside_mask(size):
    rho, theta = rho_theta(size)
    return nm_polynomial(0, 0, rho, theta, normed=False) < 1


def dict_to_list(kv):
    max_key = max(kv.keys())
    out = [0]*(max_key+1)
    for k,v in kv.items():
        out[k] = v
    return out


def ensure_dict(values, order):
    if isinstance(values,dict):
        return values
    if isinstance(values,np.ndarray):
        values = tuple(values.ravel())
    if isinstance(values,(tuple,list)):
        order = str(order).lower()
        order in ('noll','ansi') or _raise(ValueError("Could not identify the Zernike nomenclature/order"))
        offset = 1 if order=='noll' else 0
        indices = range(offset,offset+len(values))
        return dict(zip(indices,values))
    raise ValueError("Could not identify the data type for dictionary formation")



class Zernike:

    """
        Encapsulates Zernike polynomials

        :param index: string, integer or tuple, index of Zernike polynomial e.g. 'defocus', 4, (2,2)
        :param oder: string, defines the Zernike nomenclature if index is an integer, eg noll or ansi, default is noll
    """

    _ansi_names = ['piston', 'tilt', 'tip', 'oblique astigmatism', 'defocus',
                   'vertical astigmatism', 'vertical trefoil', 'vertical coma',
                   'horizontal coma', 'oblique trefoil', 'oblique quadrafoil',
                   'oblique secondary astigmatism', 'primary spherical',
                   'vertical secondary astigmatism', 'vertical quadrafoil']
    _nm_pairs = set((n,m) for n in range(200) for m in range(-n,n+1,2))
    _noll_to_nm = dict(zip((nm_to_noll(*nm) for nm in _nm_pairs),_nm_pairs))
    _ansi_to_nm = dict(zip((nm_to_ansi(*nm) for nm in _nm_pairs),_nm_pairs))

    def __init__(self, index, order='noll'):
        super().__setattr__('_mutable', True)
        if isinstance(index,str):
            if index.isdigit():
                index = int(index)
            else:
                name = index.lower()
                name in self._ansi_names or _raise(ValueError("Your input for index is string : Could not identify the name of Zernike polynomial"))
                index = self._ansi_names.index(name)
                order = 'ansi'

        if isinstance(index,(list,tuple)) and len(index)==2:
            self.n, self.m = int(index[0]), int(index[1])
            (self.n, self.m) in self._nm_pairs or _raise(ValueError("Your input for index is list/tuple : Could not identify the n,m order of Zernike polynomial"))
        elif isinstance(index,int):
            order = str(order).lower()
            order in ('noll','ansi') or _raise(ValueError("Your input for index is int : Could not identify the Zernike nomenclature/order"))
            if order == 'noll':
                index in self._noll_to_nm or _raise(ValueError("Your input for index is int and input for Zernike nomenclature is Noll: Could not identify the Zernike polynomial with this index"))
                self.n, self.m = self._noll_to_nm[index]
            elif order == 'ansi':
                index in self._ansi_to_nm or _raise(ValueError("Your input for index is int and input for Zernike nomenclature is ANSI: Could not identify the Zernike polynomial with this index"))
                self.n, self.m = self._ansi_to_nm[index]
        else:
            raise ValueError("Could not identify your index input, we accept strings, lists and tuples only")

        self.index_noll = nm_to_noll(self.n, self.m)
        self.index_ansi = nm_to_ansi(self.n, self.m)
        self.name = self._ansi_names[self.index_ansi] if self.index_ansi < len(self._ansi_names) else None
        self._mutable = False


    def polynomial(self, size, normed=True, outside=np.nan):
        
        """
            For visualization of Zernike polynomial on a disc of unit radius

            :param size: integer, Defines the shape of square grid, e.g. 256 or 512
            :param normed: boolen, Whether the Zernike polynomials are normalized, default is True
            :param outside: scalar, Outside padding of the spherical disc defined within a square grid, default is np.nan
            :return: 2D array, Zernike polynomial computed on a disc of unit radius defined within a square grid  
        """

        np.isscalar(size) and int(size) > 0 or _raise(ValueError())
        return self.phase(*rho_theta(int(size)), normed=normed, outside=outside)


    def polynomial_vertical(self, size, normed=True, outside=np.nan):  # added by srio
        np.isscalar(size) and int(size) > 0 or _raise(ValueError())

        rho = np.linspace(-1.0,1.0,size)
        w = nm_polynomial(self.n, self.m, rho, 0.0, normed=bool(normed))
        if outside is not None:
            w[nm_polynomial(0, 0, rho, 0.0, normed=False) < 1] = outside
        return w

    def phase(self, rho, theta, normed=True, outside=None):

        """
            For creation of a Zernike polynomial  with a given polar co-ordinate system

            :param rho: 2D square array,  radial axis
            :param theta: 2D square array, azimuthal axis
            :param normed: boolen, whether the Zernike polynomials are normalized, default is True
            :param outside: scalar, outside padding of the spherical disc defined within a square grid, default is None
            :return: 2D array, Zernike polynomial computed for rho and theta
        """
        (isinstance(rho,np.ndarray) and rho.ndim==2 and rho.shape[0]==rho.shape[1]) or _raise(ValueError('Only 2D square array for radial co-ordinate is accepted'))
        (isinstance(theta,np.ndarray) and theta.shape==rho.shape) or _raise(ValueError('Only 2D square array for azimutha co-ordinate is accepted'))
        size = rho.shape[0]
        np.isscalar(normed) or _raise(ValueError())
        outside is None or np.isscalar(outside) or _raise(ValueError("Only scalar constant value for outside is accepted"))
        w = nm_polynomial(self.n, self.m, rho, theta, normed=bool(normed))
        if outside is not None:
            w[nm_polynomial(0, 0, rho, theta, normed=False) < 1] = outside 

        return w


    def __hash__(self):
        return hash((self.n,self.m))


    def __eq__(self, other):
        return isinstance(other,Zernike) and (self.n,self.m) == (other.n,other.m)


    def __lt__(self, other):
        return self.index_ansi < other.index_ansi


    def __setattr__(self, *args):
        if self._mutable:
            super().__setattr__(*args)
        else:
            raise AttributeError('Zernike is immutable')


    def __repr__(self):
        return f'Zernike(n={self.n}, m={self.m: 1}, noll={self.index_noll:2}, ansi={self.index_ansi:2}' + (f", name='{self.name}')" if self.name is not None else ")")



class ZernikeWavefront:

    """
        Encapsulates the wavefront defined by Zernike polynomials

        :param amplitudes: dictionary, nd array, tuple or list, Amplitudes of Zernike polynomials
        :param oder: string, Zernike nomenclature, eg noll or ansi, default is noll 
    """
    def __init__(self, amplitudes, order='noll'):
        amplitudes = ensure_dict(amplitudes, order)
        all(np.isscalar(a) for a in amplitudes.values()) or _raise(ValueError("Could not identify scalar value for amplitudes after making a dictionary"))

        self.zernikes = {Zernike(j,order=order):a for j,a in amplitudes.items()}
        self.amplitudes_noll = tuple(dict_to_list({z.index_noll:a for z,a in self.zernikes.items()})[1:])
        self.amplitudes_ansi = tuple(dict_to_list({z.index_ansi:a for z,a in self.zernikes.items()}))
        self.amplitudes_requested = tuple(self.zernikes[k] for k in sorted(self.zernikes.keys()))


    def __len__(self):
        return len(self.zernikes)


    def polynomial(self, size, normed=True, outside=np.nan):

        """
            For visualization of weighted sum of Zernike polynomials on a disc of unit radius

            :param size: integer, Defines the shape of square grid, e.g. 256 or 512
            :param normed: boolen, Whether the Zernike polynomials are normalized, default is True
            :param outside: scalar, Outside padding of the spherical disc defined within a square grid, default is np.nan
            :return: 2D array, weighted sums of Zernike polynomials computed on a disc of unit radius defined within a square grid
        """
        return np.sum([a * z.polynomial(size=size, normed=normed, outside=outside) for z,a in self.zernikes.items()], axis=0)


    def phase(self, rho, theta, normed=True, outside=None):

        """
            For creation of phase defined as a weighted sum of Zernike polynomial with a given polar co-ordinate system

            :param rho: 2D square array,  radial axis
            :param theta: 2D square array, azimuthal axis
            :param normed: boolen, whether the Zernike polynomials are normalized, default is True
            :param outside: scalar, outside padding of the spherical disc defined within a square grid, default is none
            :return: 2D array, wavefront computed for rho and theta
        """
        return np.sum([a * z.phase(rho=rho, theta=theta, normed=normed, outside=outside) for z,a in self.zernikes.items()], axis=0)



def random_zernike_wavefront(amplitude_ranges, order='noll', rng=None):

        """
            Creates random Zernike wavefront with random amplitudes drawn from a uniform distibution

            :param aplitude_ranges: dictionary, nd array, tuple or list, amplitude bounds
            :param oder: string, to define the Zernike nomenclature if index is an integer, eg noll or ansi, default is noll
            :param rng:
            :return: Zernike wavefront object
        """
        if rng is None: rng = np.random
        amplitude_ranges = ensure_dict(amplitude_ranges, order)
        # print(">>>", amplitude_ranges, amplitude_ranges.values(), amplitude_ranges.items())


        all((np.isscalar(v) and v>=0) or (isinstance(v,(tuple,list)) and len(v)==2) for v in amplitude_ranges.values()) or _raise(ValueError())
        amplitude_ranges = {k:((-v,v) if np.isscalar(v) else v) for k,v in amplitude_ranges.items()}
        all(v[0]<=v[1] for v in amplitude_ranges.values()) or _raise(ValueError("Lower bound is expected to be less than the upper bound"))
        return ZernikeWavefront({k:rng.uniform(*v) for k,v in amplitude_ranges.items()}, order=order)


#srio
    # :param zernike_amplitude_ranges: dictionary or list, the values should either a scalar indicating the absolute magnitude
    #         or a tuple with upper and lower bound, default is {'vertical coma': (-0.2,0.2)}

if __name__ == "__main__":
    # see more examples in phasenet/notebooks/Wavefront
    import matplotlib.pylab as plt

    f = random_zernike_wavefront([(0, 0), (-1, 1), (1, 2)], order = 'ansi')
    print(f.zernikes)

    print(f.amplitudes_noll)
    print(f.amplitudes_ansi)

    plt.imshow(f.polynomial(512));
    plt.colorbar();
    plt.axis('off');
    plt.show()



import numpy as np
import matplotlib.pyplot as plt
from csbdeep.utils import _raise
# try:
#     from gputools.noise import perlin3
#     gputools_flag=False
# except:
gputools_flag=False
print(f"GPU Tools {gputools_flag}")

def normal_noise(image, mean, sigma, snr):
    noisy = np.random.normal(mean, sigma, image.shape) + image * mean * snr
    return noisy

def poisson_noise(image, snr):

    lambdaDistribution = image  * snr
    noisy = np.random.poisson(np.maximum(1, lambdaDistribution + 1).astype(int)).astype(np.float32)
    return noisy

def perlin_noise(image, mean, snr):
    Z,Y,X = image.shape
    z,y,x = np.random.normal(Z/1.5,Z/4), np.random.normal(Y/1.5,Y/4), np.random.normal(X/1.5,X/4)
    shiftz,shifty, shiftx = np.random.randint(-1000,1000), np.random.randint(-1000,1000), np.random.randint(-1000,1000)
    return image + mean + perlin3(image.shape, scale=(x,y,z),shift=(shiftx,shifty,shiftz))*mean*snr
 


def add_normal_poisson_noise(image, mean, sigma, snr):

    normal_noise_img = normal_noise(image=image, mean=mean, sigma=sigma, snr=snr)
    poisson_noise_img = poisson_noise(image=image, snr=snr)
    noisy_img = normal_noise_img + poisson_noise_img
    return noisy_img

def add_normal_poisson_perlin_noise(image, mean, sigma, snr):
    noisy_img = add_normal_poisson_noise(image=image, mean=mean, sigma=sigma, snr=snr)
    return perlin_noise(image=noisy_img, mean=mean, snr=snr)

# TODO: Make this return a function that can repeatedly been applied to an image to make it noisy
def add_random_noise(image, snr, mean, sigma, perlin=False, rng=None):
    """
        add gaussian and poisson noise to the system
        :param image: 3D array as image
        :param snr: scalar or tuple, signal to noise ratio
        :param mean: scalar or tuple, mean background noise
        :param sigma: scalar or tuple, simga for gaussian noise 
        :return: 3d array

    """
    if rng is None: rng = np.random

    snr = (snr, snr) if np.isscalar(snr) else snr
    mean = (mean, mean) if np.isscalar(mean) else mean
    sigma =(sigma, sigma) if np.isscalar(sigma) else sigma


    all(v[0]<=v[1] for v in [snr,mean,sigma]) or _raise(ValueError("Lower bound is expected to be less than the upper bound"))
    all(v[0]>=0 and v[1]>=0 for v in [snr,mean,sigma]) or _raise(ValueError("noise is expected to be greater than 0"))
    
    mean = np.random.uniform(*mean)
    sigma = np.random.uniform(*sigma) 
    snr = np.random.uniform(*snr)
    image = (image - np.min(image)) / (np.max(image) + np.min(image))
    noisy = add_normal_poisson_noise(image=image, mean=mean, sigma=sigma, snr=snr)

    if perlin:
        noisy = add_normal_poisson_perlin_noise(image=image, mean=mean, sigma=sigma, snr=snr)
    else:
        noisy = add_normal_poisson_noise(image=image, mean=mean, sigma=sigma, snr=snr)

    noisy = np.maximum(0, noisy)

    return noisy
    






from __future__ import absolute_import, print_function
from setuptools import setup, find_packages
from os import path

_dir = path.abspath(path.dirname(__file__))

with open(path.join(_dir,'phasenet','version.py'), encoding='utf-8') as f:
    exec(f.read())

with open(path.join(_dir,'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup (
    name='phasenet',
    version=__version__,
    description='PhaseNet',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mpicbg-csbd/phasenet',
    author='Debayan Saha, Martin Weigert, Uwe Schmidt',
    author_email='dsaha@mpi-cbg.de, martin.weigert@epfl.ch, uschmidt@mpi-cbg.de',
    license='BSD 3-Clause License',
    packages=find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'csbdeep>=0.4.0',
    ],
)




from mlcrl.phasenet.zernike import Zernike
import numpy
from srxraylib.plot.gol import plot, plot_table, plot_image

from oasys.util.oasys_util import write_surface_file



def write_generic_h5_surface(s, xx, yy, filename='presurface.hdf5',subgroup_name="surface_file"):
    write_surface_file(s.T, xx, yy, filename, overwrite=True)
    print("write_h5_surface: File for OASYS " + filename + " written to disk.")


def create_files_2d_zernike_sampled_profiles(nsamples,
                    noll=[6, 8, 10, 11, 12, 14, 22, 37],
                    distrubution=['n','n','n','n','n','n','n','n'],
                    scale=[1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],
                    size=128,
                    factor=1.0,
                    seed=69,
                    width=1000e-6,
                    dir="./",
                    root = "tmp_ml",
                    do_plot=False):

    # Z = numpy.zeros((size, size, nsamples))
    C = numpy.zeros((len(noll), nsamples))

    rg = numpy.random.default_rng(seed)
    x = numpy.linspace(-1, 1, size)
    y = numpy.linspace(-1, 1, size)

    xout = numpy.linspace(-width / 2, width / 2, size)
    yout = numpy.linspace(-width / 2, width / 2, size)

    # noll       = [6,     8,  10,  11,  12,  14,  22, 37]
    # distrubution = ['n', 'n', 'n', 'u', 'n', 'n', 'u', 'u']
    # scale        = [0.5, 0.5, 0.5, 2.3, .05, .05, 1.0, 0.5]

    for i in range(nsamples):
        zz = numpy.zeros((size,size))

        for ij, j in enumerate(noll):
            z = Zernike(j, order='noll')

            if distrubution[ij] == 'n':
                c = rg.normal(loc=0, scale=scale[ij] * factor)
            elif distrubution[ij] == 'u':
                c = rg.uniform(-scale[ij] * factor, scale[ij] * factor)
            else:
                raise Exception("Why am I here?")

            w = z.polynomial(size, outside=0.0)
            zz += c * w
            C[ij,i] = c
        # Z[:, :, i] = zz
        if do_plot: plot_image(zz, xout, yout, title="sampled # %s" % (i + 1))


        #
        # write file
        #

        if i == 0: # first file has no deformation
            zz * 0
            C[:,i] = 0.0
        #txt
        filename = "%s%s%06d.txt" % (dir, root, i)
        f = open(filename, 'w')

        f.write("# noll ")
        for j in range(len(noll)):
            f.write("%d  " % noll[j])
        f.write("\n")

        f.write("# coeff ")
        for j in range(len(noll)):
            f.write("%g  " % C[j,i])
        f.write("\n")

        for j in range(len(noll)):
            f.write("%d  %g  \n" % (noll[j], C[j, i]))
        f.write("\n")

        print("File written to disk: %s" % filename)

        # h5

        filename = "%s%s%06d.h5" % (dir, root, i)
        write_generic_h5_surface(zz, xout, yout, filename=filename)
        print("File written to disk: %s" % filename)

    return C






if __name__ == "__main__":



    #
    # sampled polynomials
    #
    if True:
        size = 128
        nsamples = 3
        seed = 69  # seed for generation of the random Zernike profiles
        noll         =             [6,     8,  10,  11,  12,  14,  22,  37]
        distrubution =             ['n', 'n', 'n', 'u', 'n', 'n', 'u', 'u']
        scale        = numpy.array([0.5, 0.5, 0.5, 2.3, .05, .05, 1.0, 0.5]) * 1e-6
        width = 1500e-6

        create_files_2d_zernike_sampled_profiles(nsamples,
                                                   noll=noll, distrubution=distrubution, scale=scale, factor=5.0,
                                                   size=size, seed=seed, do_plot=1,
                                                width=width)





        # x = numpy.linspace(-width/2,width/2,size)
        # y = numpy.linspace(-width / 2, width / 2, size)
        # for i in range(Z.shape[2]):
        #     plot_image(Z[:,:,i], x*1e6, y*1e6, title="sampled # %s" % (i + 1), xtitle="X [um]", ytitle="Y [um]")




from csbdeep.utils import _raise
import numpy as np
import random
import warnings
from abc import ABC, abstractmethod
import inspect
import tifffile


class Phantom3D(ABC):

    _registered = {}

    @classmethod
    def register(cls, subclass):
        issubclass(subclass, cls) or _raise(ValueError("not a subclass"))
        cls._registered[subclass.__name__.lower()] = subclass

    @classmethod
    def instantiate(cls, **kwargs):
        'name' in kwargs or _raise(ValueError("name mising"))
        name = str(kwargs['name']).lower()
        name in cls._registered or _raise(ValueError("phantom not registered"))
        subclass = cls._registered[name]
        init_keys = inspect.signature(subclass.__init__).parameters.keys()
        init_kwargs = {k:kwargs[k] for k in init_keys if k != 'name' and k in kwargs}
        return subclass(**init_kwargs)

    def __init__(self, shape):
        self.shape=shape
        len(self.shape)==3 or _raise(ValueError("Only 3D phantoms are supported"))
        self.phantom_obj = np.zeros(self.shape)

    # TODO: rename this function
    def check_phantom_obj(self):
        if np.sum(self.phantom_obj) <= 0:
            warnings.warn("No object created")

    @abstractmethod
    def generate(self):
        pass

    @abstractmethod
    def get(self):
        pass



class Points(Phantom3D):

    """
        creates multiple points
        :param obj_shape: tuple, object shape as (z,y,x), e.g. (64,64,64)
        :param num: integer, number of points, e.g. 3
        :param center: boolean, whether to have a point at the center, default is False
        :param pad_from_boundary: integer, leave space between points and bouandary. Helpful for convolution, recommended size//5
    """

    def __init__(self, shape, num, center = True, pad_from_boundary=0):
        super().__init__(shape)
        self.num = num
        self.center = center
        self.pad_from_boundary = pad_from_boundary
        self.generate()

    def generate(self):
        _num = self.num
        np.isscalar(self.pad_from_boundary) and self.pad_from_boundary<np.min(self.shape)//2  or _raise(ValueError( "Padding from boundary has to be scalar and bouded by object size"))
        
        x = np.zeros(self.shape, np.float32)
        if self.center:
            x[self.shape[0] // 2, self.shape[1] // 2, self.shape[2] // 2] = 1
            _num = _num-1
        _i, _j, _k = np.random.randint(self.pad_from_boundary, np.min(self.shape) - self.pad_from_boundary, (3,_num))
        x[_i, _j, _k] = 1.

        self.phantom_obj = x
        self.check_phantom_obj()

    def get(self):
        self.check_phantom_obj()
        return self.phantom_obj



class Sphere(Phantom3D):

    """
        creates 3D sphere
        :param obj_shape: tuple, object shape as (z,y,x), e.g. (64,64,64)
        :param units: tuple, voxel size in microns, e.g. (0.1,0.1,0.1)
        :param radius: scalar, radius of sphere in microns, e.g. 0.5
        :param off_centered: tuple, displacement vector by which center is moved as (k,j,i) e.g. (0,0.5,0.5)
    """

    def __init__(self, shape, units, radius, off_centered=(0,0,0)):
        super().__init__(shape)
        self.shape=shape
        self.units= units
        self.radius = radius
        self.off_centered = off_centered
        self._img = np.zeros(shape)
        self.generate()

    def generate(self):

        isinstance(self.off_centered,tuple) or _raise(ValueError("Displacement vector for center is not a 3D vector"))
        if isinstance(self.radius,(list,tuple)):
            self.radius = random.choice(self.radius)
        np.isscalar(self.radius) or _raise(ValueError("Radius has to be scalar"))
        all(2*self.radius<_u*_o for _u,_o in zip(self.units,self.shape)) or _raise(ValueError("Object diameter is bigger than object shape"))

                
        xs = list(u*(np.arange(s)-s/2) for u,s in zip(self.units, self.shape)) 
        xs = tuple(_x-self.off_centered[i] for i,_x in enumerate(xs))
        Z,Y,X = np.meshgrid(*xs,indexing = "ij")
        R = np.sqrt(X**2+Y**2+Z**2)
        mask = 1.*(R<=self.radius)

        self.phantom_obj = mask
        self.check_phantom_obj()
    
    def get(self):
        self.check_phantom_obj()
        return self.phantom_obj

class Images(Phantom3D):

    """
        creates 3D sphere
        :param obj_shape: tuple, object shape as (z,y,x), e.g. (64,64,64)
        :param units: tuple, voxel size in microns, e.g. (0.1,0.1,0.1)
        :param radius: scalar, radius of sphere in microns, e.g. 0.5
        :param off_centered: tuple, displacement vector by which center is moved as (k,j,i) e.g. (0,0.5,0.5)
    """

    def __init__(self, shape, filepath, augment=False):
        # TODO : augment
        super().__init__(shape)
        self.shape=shape
        self.augment = augment
        self.image = self.get_image(filepath)
        self.generate()

    def get_image(self, filepath):
        return tifffile.imread(filepath)

    def generate(self):

        self.image.ndim == 3 or _raise(ValueError("3D image required"))

        # TODO : augment, crop/pad according to shape
        self.phantom_obj = self.image
        self.check_phantom_obj()
    
    def get(self):
        self.check_phantom_obj()
        return self.phantom_obj

Phantom3D.register(Points)
Phantom3D.register(Sphere)
Phantom3D.register(Images)



from mlcrl.phasenet.zernike import Zernike
import numpy
from srxraylib.plot.gol import plot, plot_table, plot_image
import matplotlib.pyplot as plt

def create_2d_zernike_sampled_profiles(nsamples,
                    noll=[6, 8, 10, 11, 12, 14, 22, 37],
                    distrubution=['n','n','n','n','n','n','n','n'],
                    scale=[1.0,1.0,1.0,1.0,1.0,1.0,1.0,1.0],
                    size=128,
                    factor=1.0,
                    seed=69,
                    do_plot=False):

    Z = numpy.zeros((size, size, nsamples))
    C = numpy.zeros((len(noll), nsamples))

    rg = numpy.random.default_rng(seed)
    x = numpy.linspace(-1, 1, size)
    y = numpy.linspace(-1, 1, size)

    # noll       = [6,     8,  10,  11,  12,  14,  22, 37]
    # distrubution = ['n', 'n', 'n', 'u', 'n', 'n', 'u', 'u']
    # scale        = [0.5, 0.5, 0.5, 2.3, .05, .05, 1.0, 0.5]

    for i in range(nsamples):
        zz = numpy.zeros((size,size))

        for ij, j in enumerate(noll):
            z = Zernike(j, order='noll')

            if distrubution[ij] == 'n':
                c = rg.normal(loc=0, scale=scale[ij] * factor)
            elif distrubution[ij] == 'u':
                c = rg.uniform(-scale[ij] * factor, scale[ij] * factor)
            else:
                raise Exception("Why am I here?")

#Generator(PCG64) 0 1.785774589511436e-06 7
# Generator(PCG64) 0 8.465043621172215e-07 7
# Generator(PCG64) 0 6.654259141073615e-06 7
# Generator(PCG64) 0 5.3994894946457855e-06 7
# Generator(PCG64) 0 -3.987190882967194e-07 7
# Generator(PCG64) 0 3.166962286232809e-06 7
# Generator(PCG64) 0 2.231738805778543e-07 7
# Generator(PCG64) 1 -1.3682866656138383e-06 7
# Generator(PCG64) 1 -2.0353079624572042e-07 7
# Generator(PCG64) 1 -8.597814005268755e-07 7
# Generator(PCG64) 1 6.585212425935185e-06 7
# Generator(PCG64) 1 -1.0695773383463991e-07 7
# Generator(PCG64) 1 4.621093285144029e-06 7
#             print(rg,i,c, len(noll))
            w = z.polynomial(size, outside=0.0)
            zz += c * w
            C[ij, i] = c
        Z[:, :, i] = zz

        if do_plot: plot_image(zz, x, y, title="sampled # %s" % (i + 1))


    return C, Z

if __name__ == "__main__":



    # For visualizing the Zernikes we define the Zernike object and call `polynomial()`.
    # The size of the 2D array is passed as a parameter

    if False:
        fig, ax = plt.subplots(4,10, figsize=(16,10))
        z_old = Zernike(1, order='noll').polynomial(128)
        for i,a in enumerate(ax.ravel()):
            z = Zernike(i+1, order='noll')
            w = z.polynomial(128)
            print(w.shape)
            a.imshow(w)
            if z.name is not None:
                a.set_title(z.name)
                # a.set_title("Noll: %d (%d,%d)" % (i + 1, z.n, z.m))
            else:
                a.set_title("Noll: %d (%d,%d)" % (i+1, z.n, z.m))
            a.axis('off')
            # print(">>>>>> i.(i-1) = ", numpy.nansum((z_old * w)) )

        plt.show()


    #
    # check normalization 2D
    #
    if False:
        noll = [6, 8, 10, 11, 12, 14, 22, 37]
        nnoll = len(noll)
        size = 128 * 4
        for i in range(nnoll):
            for j in range(nnoll):
                z1 = Zernike(noll[i], order='noll')
                z2 = Zernike(noll[j], order='noll')
                tmp = numpy.nansum(z1.polynomial(size=size) * z2.polynomial(size=size))/size**2
                if numpy.abs(tmp) < 1e-12:
                    tmp = 0
                print(">>> norm noll %d  %d: %g" % (noll[i],noll[j], tmp))




    #
    # sampled polynomials
    #
    if True:
        size = 128
        nsamples = 16
        seed = 69  # seed for generation of the random Zernike profiles
        noll         =             [6,     8,  10,  11,  12,  14,  22,  37]
        distrubution =             ['n', 'n', 'n', 'u', 'n', 'n', 'u', 'u']
        scale        = numpy.array([0.5, 0.5, 0.5, 2.3, .05, .05, 1.0, 0.5]) * 1e-6
        width = 1500e-6

        C, Z  = create_2d_zernike_sampled_profiles(nsamples,
                                                   noll=noll, distrubution=distrubution, scale=scale, factor=5.0,
                                                   size=size, seed=seed, do_plot=0)


        print(Z.shape,C)


        x = numpy.linspace(-width/2,width/2,size)
        y = numpy.linspace(-width / 2, width / 2, size)
        for i in range(Z.shape[2]):
            plot_image(Z[:,:,i], x*1e6, y*1e6, title="sampled # %s" % (i + 1), xtitle="X [um]", ytitle="Y [um]")

        #
        # write files
        #
        # dir = "./"
        # root = "tmp_ml"
        #
        # for i in range(nsamples):
        #     #txt
        #     filename = "%s%s%04d.txt" % (dir, root, i+1)
        #     f = open(filename, 'w')
        #
        #     f.write("# noll ")
        #     for j in range(len(noll)):
        #         f.write("%d  " % noll[j])
        #     f.write("\n")
        #
        #     f.write("# coeff ")
        #     for j in range(len(noll)):
        #         f.write("%g  " % C[j,i])
        #     f.write("\n")
        #
        #     print("File written to disk: %s" % filename)
        #
        #     # dat
        #     filename = "%s%s%04d.dat" % (dir, root, i + 1)
        #     f = open(filename, 'w')
        #
        #     for j in range(size):
        #         f.write("%g   %g\n" % (xx[j], Y[j,i]))
        #     f.close()
        #     print("File written to disk: %s" % filename)
        #
        #
        # # overwrite 000 with no deformation
        # # txt - no deformation
        # filename = "%s%s%04d.txt" % (dir, root, 0)
        # f = open(filename, 'w')
        #
        # f.write("# noll ")
        # for j in range(len(noll)):
        #     f.write("%d  " % noll[j])
        # f.write("\n")
        #
        # f.write("# coeff ")
        # for j in range(len(noll)):
        #     f.write("%g  " % 0.0)
        # f.write("\n")
        #
        # print("File written to disk: %s" % filename)
        #
        # # dat - no deformation
        # filename = "%s%s%04d.dat" % (dir, root, 0)
        # f = open(filename, 'w')
        #
        # for j in range(size):
        #     f.write("%g   %g\n" % (xx[j], 0.0))
        # f.close()
        # print("File written to disk: %s" % filename)



from csbdeep.utils import _raise
import random

def cropper3D(image, crop_shape, jitter=False, max_jitter=None, planes=None):

	"""
        crops 3d images
        
        :param image: 3d array, image to be cropped
        :param crop_shape: tuple, crop shape
        :param jitter: booelan, randomly move the center point within a given limit, default is False
        :param max_jitter: tuple, maximum displacement for jittering, if None then it gets a default value
	"""

	if planes is None:
		half_crop_shape = tuple(c//2 for c in crop_shape)
	else:
		half_crop_shape = tuple((crop_shape[-1]//2,crop_shape[-1]//2,crop_shape[-1]//2))

	half_image_shape = tuple(i//2 for i in image.shape)
	assert all([c<=l for c,l in zip(half_crop_shape,half_image_shape)]), "Crop shape is bigger than image shape"

	if jitter:
		contraint_1 = tuple((l-c)//4 for c,l in zip(half_crop_shape,half_image_shape))
		contraint_2 = tuple(c//2 for c in half_crop_shape)

		if max_jitter is None:
			max_jitter = tuple([min(_x,_y) for _x,_y in zip(contraint_1,contraint_2)])
		assert all([l-m>=0 and l+m<2*l for m,l in zip(max_jitter,half_image_shape)]), "Jittering results in cropping outside border, please reduce max_jitter"
		loc = tuple(l-random.randint(-1*max_jitter[i],max_jitter[i]) for i,l in enumerate(half_image_shape))
	else:
		loc=half_image_shape

	final_img = image[loc[0]-half_crop_shape[0]:loc[0]+half_crop_shape[0], loc[1]-half_crop_shape[1]:loc[1]+half_crop_shape[1], loc[2]-half_crop_shape[2]:loc[2]+half_crop_shape[2]]

	if planes is not None:
		try:
			final_img = final_img[planes+half_crop_shape[0]]
		except:
			_raise(ValueError("Plane does not exist"))

	return final_img 

